;; Generated package description from matlab-mode.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "matlab-mode" "8.2.0.0.20260428.134050" "Major mode for MATLAB(R) dot-m files" '((emacs "27.2")) :commit "83ea45983975bc6d17a908bf4acddfc47d51763a" :authors '(("Matt Wette" . "mwette@alumni.caltech.edu") ("Eric M. Ludlam" . "eludlam@mathworks.com")) :maintainer '(("Eric M. Ludlam" . "eludlam@mathworks.com") ("Uwe Brauer" . "oub@mat.ucm.es") ("John Ciolfi" . "john.ciolfi.32@gmail.com")) :keywords '("matlab(r)") :url "https://github.com/mathworks/Emacs-MATLAB-Mode")
